/*******************************************************************************
 * Copyright (c) 2001 IBM Corporation.
 * All rights reserved. This program and the accompanying materials 
 * are made available under the terms of the Common Public License v0.5 
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v05.html
 * 
 * Contributors:
 *     IBM Corporation - example implementation
 ******************************************************************************/

package com.example;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IFolder;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IResourceVisitor;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.IPackageDeclaration;
import org.eclipse.jdt.core.IPackageFragment;
import org.eclipse.jdt.core.IPackageFragmentRoot;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.JavaModelException;

/**
 * Minimize the number of loose class files from an unzipped class library
 * by removing ones corresponding to Java source files.
 */
public class MinimizeLib {
	
	static boolean DEBUG = true;

	/**
	 * Example of discarding class files for which there is corresponding 
	 * source.
	 * <p>
	 * Assumptions:
	 * <ul>
	 * <li>The source files are in the "src" root folder of the project.</li>
	 * <li>The "src" folder is a source package fragment root. This allows
	 *    us to use the Java model to find and access source files.</li>
	 * <li>The class files are in the "classes" root folder of the project.</li>
	 * <li>We want to delete any class files in the "classes" folder for which
	 *    there is corresponding source file in the "src" folder.</li>
	 * </ul>
	 * </p>
	 * 
	 * @param project the Java project to minimize
	 */
	public static void minimize(IJavaProject project) {

		if (DEBUG) {
			System.out.println("Minimizing project " + project.getElementName());
		}

		// Project/src contains source files
		// Project/classes contains class files
		final IFolder srcFolder = project.getProject().getFolder("src");
		final IFolder classesFolder = project.getProject().getFolder("classes");
		
		if (!srcFolder.exists()) {
			// no source folder means no pruning possible
			if (DEBUG) {
				System.out.println("No source folder " + srcFolder.getFullPath());
			}
			return;
		}
		if (!classesFolder.exists()) {
			// no classes folder means nothing to prune
			if (DEBUG) {
				System.out.println("No library folder " + classesFolder.getFullPath());
			}
			return;
		}
		
		// List of fully qualified type names for which we have source
		// (element type: String)
		final Set sourceTypeNames = new HashSet(1000);
		
		// use Java model to rip through sources to get list of type names
		try {
			IPackageFragmentRoot srcRoot = project.getPackageFragmentRoot(srcFolder);
			IJavaElement[] pkgs = srcRoot.getChildren();
			for (int i=0; i<pkgs.length; i++) {
				if (pkgs[i] instanceof IPackageFragment) {
					IPackageFragment pkg = (IPackageFragment) pkgs[i];
					ICompilationUnit[] cus = pkg.getCompilationUnits();
					for (int j=0; j<cus.length; j++) {
						ICompilationUnit cu = cus[j];
						sourceTypeNames.addAll(Arrays.asList(extractTypeNames(cu)));
					}
				}
			}
		} catch (JavaModelException e) {
			// unexpected
			e.printStackTrace();
		}
		
		if (sourceTypeNames.isEmpty()) {
			// no source types means no pruning possible
			if (DEBUG) {
				System.out.println("No source types");
			}
			return;
		}

		// verify that none of the type names are problematic (contain "$")
		// if the name contains '$' we can't pattern match class file names
		for (Iterator it = sourceTypeNames.iterator(); it.hasNext(); ) {
			String sourceTypeName = (String) it.next();
			if (sourceTypeName.indexOf('$') >= 0) {
				// we're in trouble
				throw new RuntimeException("Some source types have '$' in their name");
			}
		}

		// walk the classes folder deleting class files for which there is source
		class Visitor implements IResourceVisitor {
			public boolean visit(IResource res) throws CoreException {
				if (res.getType() == IResource.FILE) {
					IFile file = (IFile) res;
					String ext = res.getFileExtension();
					if (ext != null && ext.equals("class")) {
						IPath pkgPath = file.getFullPath().removeFirstSegments(2).removeLastSegments(1);
						String pkgName = pkgPath.toString().replace('/', '.');
						String baseTypeName = 
							pkgName.length() == 0
								? baseTypeName(file)
								: pkgName + "." + baseTypeName(file);
						if (DEBUG) {
							System.out.println("Have source for " + baseTypeName 
								+ "? " + (sourceTypeNames.contains(baseTypeName) ? "Yes" : "No"));
						}
						if (sourceTypeNames.contains(baseTypeName)) {
							deleteClassFile(file);
						}
					}
				}
				return true;
			}
		}

		try {
			classesFolder .accept(new Visitor());
		} catch (CoreException e) {
			// should not happen
			e.printStackTrace();
		}
	}
		
	/**
	 * Deletes the given class file resource. Does nothing if unable to
	 * delete it.
	 * 
	 * @param classFile the class file resource to delete
	 */
	static void deleteClassFile(IFile classFile) {
		if (DEBUG) {
			System.out.println("Delete " + classFile.getFullPath());
		}
		try {
			classFile.delete(true, false, (IProgressMonitor) null);
		} catch (CoreException e) {
			// unexpected
			e.printStackTrace();
		}
	}

	/**
	 * Returns the base type name for the given class file, assuming
	 * the standard scheme Java compilers use to name generated class files.
	 * <p>
	 * For example,
	 * <ul>
	 * <li>file <code>Foo.class</code> returns <code>"Foo"</code></li>
	 * <li>file <code>Foo$1.class</code> returns <code>"Foo"</code></li>
	 * <li>file <code>Foo$Bar.class</code> returns <code>"Foo"</code></li>
	 * <li>file <code>Foo$1$Bar.class</code> returns <code>"Foo"</code></li>
	 * </ul>
	 * 
	 * @param classFile the class file
	 * @return the name of the corresponding top-level type
	 */	
	static String baseTypeName(IFile classFile) {
		String fileName = classFile.getName();
		int x = fileName.lastIndexOf(".class");
		if (x < 0) {
			throw new IllegalArgumentException();
		}
		// strip off .class suffix
		String binaryTypeName = fileName.substring(0, x);
		int d = binaryTypeName.indexOf("$");
		if (d < 0) {
			return binaryTypeName;
		} else {
			// the characters before the '$' is the top-level type name
			return binaryTypeName.substring(0, d);
		}
	}

	/**
	 * Returns a list of fully-qualifed names of top-level types declared
	 * in the given compilation unit. Returns the empty list if the compilation
	 * unit is empty or could not be parsed.
	 * 
	 * @param cu the Java compilation unit
	 * @return the list of fully-qualified names of package-member types
	 */
	static String[] extractTypeNames(ICompilationUnit cu) {
		List typeNames = new ArrayList();
		try {
			IPackageDeclaration[] pds = cu.getPackageDeclarations();
			String packageName = pds.length == 0 ? "" : pds[0].getElementName();
			IType[] types = cu.getTypes();
			for (int k =0; k<types.length; k++) {
				IType type = types[k];
				String name = type.getElementName();
				String fqTypeName = 
					packageName.length() == 0 ? name : packageName + "." + name;
				typeNames.add(fqTypeName);
			}
		} catch (JavaModelException e) {
			// unexpected
			e.printStackTrace();
		}
		String[] result = new String[typeNames.size()];
		typeNames.toArray(result);
		return result;
	}

}